import { is } from '../fn.js'

class Value {
  #value

  constructor(value) {
    this.#value = value
  }

  get value() {
    return this.#value?.value ?? this.#value
  }

  equals(other) {
    if (other == null) {
      return false
    }

    const value = this.value
    return value === other
      || value === other.value
  }
}

export class Identifier extends Value {
  constructor(value) {
    super(value)
  }
}

export class Name extends Identifier {
  constructor(value) {
    super(value)
  }
}

export class Label extends Identifier {
  constructor(value) {
    super(value)
  }
}

export class SelectedGraph extends Identifier {
  constructor(value) {
    super(value)
  }
}

export class Literal {
  constructor(value) {
    this.value = value
  }
}


export class Alias extends Identifier {
  constructor(value, alias) {
    super(value)
    this.alias = alias
  }
}

export class ObjectPath extends Identifier {
  constructor(value, ...path) {
    super(value)
    this.path = path
  }

  toString() {
    return this.path.reduce((acc, val) => acc + val, this.value)
  }

  from(obj) {
    return this.path.reduce((obj, key) => obj && obj[key], obj)
  }

  equals(value) {
    if (is(ObjectPath, value)) {
      return value != null
        && this.path.length === value.path.length
        && super.equals(value)
        && this.path.every((x, i) => x === value.path[i])
    } else {
      return super.equals(value)
    }
  }
}


class GraphObject {
  constructor(name, label, properties = []) {
    this.name = name
    this.label = label
    this.properties = properties
  }
}

export class Component extends GraphObject {
  constructor(name, label, properties) {
    super(name, label, properties)
  }
}

export class Node {
  constructor(components) {
    this.components = components
  }
}


export const Direction = Object.freeze({
  Left: 0,
  Right: 1
})

export class Edge {
  constructor(components) {
    this.components = components
  }
}

export class DirectedEdge extends Edge {
  constructor(components, direction) {
    super(components)
    this.direction = direction
  }

  static fromEdge(edge, direction) {
    return new DirectedEdge(edge.components, direction)
  }
}

export class Relationship {
  constructor(left, edge, right) {
    const [from, to] = edge.direction === Direction.Right ? [left, right] : [right, left]
    this.from = from
    this.to = to
    this.edge = edge
  }
}

export class ReturnValues {
  constructor(...args) {
    this.values = args
  }

  find(fn) {
    return this.values.find(fn)
  }

  includes(value) {
    return this.find(x => x.equals(value))
  }
}

export class KeyValuePair {
  constructor(key, value) {
    this.key = key
    this.value = value
  }
}

export class Query {
  constructor(use, match, returnValues) {
    this.use = use
    this.match = match
    this.returnValues = returnValues
  }
}
